// senc.cpp

// $Source: /u/cvs/senc/senc.cpp,v $
// $Revision: 2.5 $
// $Author: heinz $
// $Date: 2003/05/29 18:43:27 $
// $Name:  $
// $State: Exp $

/////////////////////////////////////////////////////////////////////////////
//
// This program has been written by
//
//        Heinz Spiess, EMME/2 Support Center, CH-2558 Aegerten, Switzerland
//        heinz@spiess.ch
//
// 
// Compiled binaries of this program can be distributed under the following
// conditions:
//
//  1) This program, as well as programs derived from it, cannot be sold
//     for a fee (but it can be used and distributed without charge within
//     consulting projects).
//
//  2) The user is aware that this program is not a part of the EMME/2 .
//     Enid software licence and there is no explicit or implied warranty
//     or support provided with this program.
//
// In addition to the above, the author grants the right to access and
// using the source code of this program as an example application for
// writing Enif client programs.  Integrating parts of the source code into
// other programs or making modifications to the programs is permitted
// under the following additional conditions:
//
//  3) Anyone who changes, enhances or otherwise modifies the source code
//     of SEnC is obliged to rename the changed source files and to use
//     a different program name for the modified program.
//
//  4) The source code of such derived programs must contain a reference
//     to the SEnC program and the particular version which it based on.
//
//						Heinz Spiess,
//						Aegerten, 24. Mai 2003
//
/////////////////////////////////////////////////////////////////////////////

//  Things still to do:
//  - improve search for available port
//  - improved error handling and reporting
//  - enhance interactive window to allow file opening 
//

// $Log: senc.cpp,v $
// Revision 2.5  2003/05/29 18:43:27  heinz
// avoid using cerr, use a QTextStream instead
//
// Revision 2.4  2003/05/29 09:17:39  heinz
// improved search for location of Enif binary with option -l
//
// Revision 2.3  2003/05/28 20:01:29  heinz
// improved implementation of the prompt client command
//
// Revision 2.2  2003/05/28 18:35:26  heinz
// the version identifier is now extracted from the CVS keys
//
// Revision 2.1  2003/05/28 18:27:33  heinz
// version corresponding the revised SEnC paper dated May 2003
//

#include <qglobal.h>
#ifdef Q_OS_WIN32
#include <io.h>
#else
#include <unistd.h>
#endif // Q_OS_WIN32
#include <stdlib.h>

#include <qtextstream.h>
#include <qmenubar.h>
#include <qsocket.h>
#include <qregexp.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qdns.h>
#include <qtimer.h>
#include <qapplication.h>
#include <qevent.h>
#include <qmessagebox.h>
#include <qvbox.h>
#include <qprocess.h>

#include "senc.h"

int Debug = 0;

bool Verbose = FALSE;
bool Interactive = FALSE;
bool EchoOnTerminal = TRUE;
bool StopOnKO = TRUE;
bool QuitOnEof = FALSE;
bool LaunchEnif = FALSE;

unsigned int EnifPort = 0;
QProcess *EnifProcess = 0;

QString EnifHost("localhost");
QString OK("OK");
QString KO("KO");
QFile OutputFile;
QTextStream *OutStream;
QTextStream ErrStream(stderr,IO_WriteOnly|IO_Translate);
QString Version;

Client *MyClient;

QRegExp Client::validName("[A-Za-z][A-Za-z0-9_]*");

////////////////////////////////////////////////////////////////////////////

Client::Client( const QString& host, unsigned int port, unsigned int delay):
    	Args(),Command(),CmdFile(),Timer()
{
    Port = port;
    Host = host;
    TextStream = 0;
    CmdStream = 0;
    CmdLine = 0;
    Stopped = FALSE;
    Waiting = FALSE;
    MaxStartupDelay = delay;
    DelayIncrement = delay<100?delay:100;
    Socket = new QSocket( this );
    connect ( &Timer, SIGNAL( timeout() ), this, SLOT( resume() ) );
    connect ( Socket, SIGNAL( error(int) ), this, SLOT( socket_error(int) ) );
    connect ( Socket, SIGNAL(connectionClosed()), this, SLOT(terminate()));
    connect ( Socket, SIGNAL( readyRead() ), this, SLOT( readyRead() ) );
    connect ( Socket, SIGNAL( hostFound() ), this, SLOT( hostFound() ) );
    connect ( Socket, SIGNAL( connected() ), this, SLOT( connected() ) );
    QTimer::singleShot ( DelayIncrement, this, SLOT(connectToHost())); 
}

////////////////////////////////////////////////////////////////////////////

Client::~Client()
{
    delete Socket;
}

////////////////////////////////////////////////////////////////////////////

void Client::hostFound()
{
    if(Debug) ErrStream << tr( "host found %1\n" ).arg( Host );
}

////////////////////////////////////////////////////////////////////////////

void Client::connectToHost()
{
    static bool firsttime = TRUE;
    MaxStartupDelay -= DelayIncrement;
    if(firsttime && Verbose)
	ErrStream <<tr("Connecting to host %1 on port %2.\n").arg(Host).arg(Port);
    Socket->connectToHost(Host,Port);
    if(firsttime && Verbose)
	ErrStream <<tr("Socket created: status=%1\n").arg(Socket->status());
    firsttime = FALSE;
}

////////////////////////////////////////////////////////////////////////////

void Client::connected()
{
    if(Verbose) ErrStream << tr( "Connected to %1.\n" ).arg( Socket->peerName() );
    TextStream = new QTextStream(Socket);
}

////////////////////////////////////////////////////////////////////////////

void Client::socket_error(int ec)
{
    switch(ec){
    case 0:
	if(MaxStartupDelay>0){
	    QTimer::singleShot ( DelayIncrement, this, SLOT(connectToHost()));
	    return;
	}
	ErrStream <<tr("No Enif server found on port %1 of host %2!\n")
	    .arg(Port).arg(Host);
 	break;
    default:
        ErrStream <<"error "<<ec<<"!\n";
    }
    terminateAll(ec);
}

////////////////////////////////////////////////////////////////////////////

void Client::readyRead()
{
    if ( !Socket->canReadLine() )
	return;

    QString response;
    bool okko = FALSE;
    do {
        response = Socket->readLine();
	emit receivedText(response);
	response.truncate(response.length()-2);
        if(response == KO){
	    if(Debug)ErrStream <<
		tr("Negative server response at line %1 of file %2!\n")
		    .arg(CmdLine).arg(CmdFile.name());
	}
        okko = (response==OK || response==KO);
        if(okko){
	    KOReceived = (response==KO);
	    okko = !sendCommand();
        }else{
            if(Variables.contains("{OUTPUT}")){
	        Variables["{OUTPUT}"] = response;
	    }
            if(!Variables.contains("{NOECHO}")){
	        *OutStream << response<<"\n";
	        OutStream->device()->flush();
	    }
	}
    } while( Socket->canReadLine() );

    if(okko && QuitOnEof && Command.isEmpty() && !Waiting)terminate();
}

////////////////////////////////////////////////////////////////////////////

void Client::insertCommand(const QString & newcmd)
{
    if(newcmd.isEmpty())return;
    Stopped = FALSE;
    Command = QString();
    CmdStack.prepend(newcmd);
    checkCommand();
    if(!Command.isEmpty())sendCommand();
}

////////////////////////////////////////////////////////////////////////////

void Client::sendText(const QString & text)
{
    if(text.isEmpty())return;
    Socket->writeBlock(text.latin1(),text.length());
}

////////////////////////////////////////////////////////////////////////////

void Client::terminate()
{
    terminateAll();
}

////////////////////////////////////////////////////////////////////////////

void Client::deleteMe()
{
    delete this;
}

////////////////////////////////////////////////////////////////////////////

void Client::setArgs(const QStringList& fl)
{
    Args = fl;
    checkCommand();
}

////////////////////////////////////////////////////////////////////////////

void Client::addArg(const QString& fn)
{
    Args << fn;
    checkCommand();
}

////////////////////////////////////////////////////////////////////////////

bool Client::cmdAvailable()
{
    if(Waiting || Stopped) return FALSE;
    checkCommand();
    return !Command.isEmpty();
}

////////////////////////////////////////////////////////////////////////////

bool Client::sendCommand()
{
    if(cmdAvailable()){
	if(StopOnKO && KOReceived){
	    Stopped = TRUE;
	    return FALSE;
        }
        QString message = getCommand(); 
        sendText(message+"\r\n");
        if(EchoOnTerminal) emit sentText(message,"COLOR=#0000FF");
	return TRUE;
    }
    return FALSE;
}

////////////////////////////////////////////////////////////////////////////

void Client::checkCommand()
{
    if(Waiting || Stopped || !Command.isNull())return;
    if(!CmdStack.count() && (!CmdStream || CmdStream->atEnd())){
	if(CmdStream)delete CmdStream;
	CmdStream = 0;
	CmdFile.close();
        if(Args.count()==0)return;
	while(Args.count()>0){
	    QString spec = Args[0];
	    Args.pop_front();
	    if(Debug>1)ErrStream <<Args.count()<<" "<<spec<<"\n";
	    int sepeq = spec.find('=');
	    if(sepeq<0){ // it's a file name
	        CmdFile.setName(spec);
		break;
            }else{               // it's a variable specification
		setVariable(spec);
		if(Args.count()==0)return;
	    }          
        }
        if(CmdFile.name()=="-"){
            CmdStream = new QTextStream(stdin,IO_ReadOnly);
	    if(Variables.contains("{FILENAME}"))
	        Variables["{FILENAME}"] = "-";
	}else{
	    if(CmdFile.open(IO_ReadOnly|IO_Translate)){
		setLineNumber(0);
		CmdLine = 0;
		if(Variables.contains("{LINENUMBER}"))
		    Variables["{LINENUMBER}"] = QString("%1").arg(CmdLine);
	        CmdFile.at(0);
	        if(Verbose)
		    ErrStream << tr("File %1 opened.\n").arg(CmdFile.name());
	        CmdStream = new QTextStream(&CmdFile);
		if(Variables.contains("{FILENAME}"))
		    Variables["{FILENAME}"] = CmdFile.name();
	    }else{
	        error(tr("Cannot open file %1!\n").arg(CmdFile.name()),1);
	    }
        }
    }
    while(CmdStack.count() || (CmdStream && !CmdStream->atEnd())){
	if(CmdStack.count()){
	    Command = CmdStack[0];
	    CmdStack.pop_front();
	}else{
	    Command = CmdStream->readLine();
	    setLineNumber(CmdLine+1);
	}
	if(Debug>1)ErrStream <<tr("reading: %1\n").arg(Command);
	if(Variables.count()>0 && Command.contains('{'))
	    Command = substituteVariables(Command);
	if(!clientCommand(Command))return;
	if(Waiting || Stopped) break;
    }
    Command=QString();
    if(!(Waiting || Stopped))checkCommand();
}

////////////////////////////////////////////////////////////////////////////

void Client::setLineNumber(int lineno)
{
    CmdLine = lineno;
    if(Variables.contains("{LINENUMBER}"))
        Variables["{LINENUMBER}"] = QString("%1").arg(CmdLine);
}

////////////////////////////////////////////////////////////////////////////

bool Client::clientCommand(const QString& cmd)
{
    if(cmd.isEmpty() || cmd[0]=='#')return TRUE;
    if(cmd[0]!='!')return FALSE;

    emit sentText(cmd,"COLOR=#00FF00");
    QStringList tl = QStringList::split(' ',cmd.mid(1).simplifyWhiteSpace());
 
    int count = tl.count();
    if(count==0)return TRUE;
    if(Debug)ErrStream <<tr("client command: %1\n").arg(cmd);
    if(tl[0].lower()=="wait"){                   // !wait <seconds>
	Waiting = TRUE;
	Timer.start(count>1?int(tl[1].toFloat()*1000):1000,TRUE);
    }else if(tl[0].lower()=="function"){           // !function <fct> <arg> ...
	QStringList function;
	function << cmd.mid(10);
        while(!CmdStream->atEnd()){
	    Command = CmdStream->readLine();
	    if(Command.lower() == "!!") break; 
	    function << Command;
	}
	if(count>1){
	    Functions[tl[1]] = function;
	    if(Debug)ErrStream << tr("Function %1 defined.\n").arg(tl[1]);
	}
    }else if(tl[0].lower()=="if"){               // if <v1> <comp> <v1>
	bool condition = FALSE;
	if(count==2){
	    condition = (tl[1].toFloat()!=0) || (tl[1].lower()=="yes");
	}else if(count>2){
	    QString s1 = tl[1];
	    double v1 = s1.toDouble();
            QString s2("");
	    double v2 = 0;
	    if(count>3){
		s2 = tl[3];
	        v2 = s2.toDouble();
	    }
	    if(tl[2]=="=="){ // string equal
                condition = (s1==s2);
	    }else if(tl[2]=="!="){ // string not equal
                condition = (s1!=s2);
	    }else if(tl[2]=="="){ // numerically equal
                condition = (v1==v2);
	    }else if(tl[2]==">"){ // numerically greater than
                condition = (v1>v2);
	    }else if(tl[2]=="<"){ // numerically less than
                condition = (v1<v2);
	    }else if(tl[2]==">="){ // numerically greater or equal
                condition = (v1>=v2);
	    }else if(tl[2]=="<="){ // numerically less of equal
                condition = (v1<=v2);
	    }else if(tl[2]=="<>"){ // numerically not equal
                condition = (v1!=v2);
	    }
	}
	if(!condition){
	    if(CmdStack.count()){
	        CmdStack.pop_front();
	    }else{
	        if(!CmdStream->atEnd()){
	            (void) CmdStream->readLine();
	            setLineNumber(CmdLine+1);
		}
	    }
	}
    }else if(tl[0].lower()=="iferror"){     // iferror
	if(!KOReceived){
	    if(CmdStack.count()){
	        CmdStack.pop_front();
	    }else{
	        if(!CmdStream->atEnd()){
	            (void) CmdStream->readLine();
	            setLineNumber(CmdLine+1);
		}
	    }
	}
	KOReceived = FALSE;
    }else if(tl[0].lower()=="for"){ // !for <var> <v1> <v2> <inc>
	if(count>3){
	    QString var = tl[1];
	    if(validName.exactMatch(var)){
		 double startval = tl[2].toDouble();
		 double endval = tl[3].toDouble();
		 double incval = count>4?tl[4].toDouble():1;
		 QString newcmd;
	         if(CmdStack.count()){
		     newcmd = CmdStack[0];
	             CmdStack.pop_front();
	         }else{
	             if(!CmdStream->atEnd()){
	                 newcmd = CmdStream->readLine();
	                 setLineNumber(CmdLine+1);
		     }
		 }
                 if( (incval > 0 && startval <= endval) ||
		     (incval < 0 && startval >= endval)){
		     if(!newcmd.isEmpty()){
			 Variables["{"+var+"}"] = tl[2];
		         QString newfor = 
			     QString("!for %1 %2 %3 %4")
			     .arg(var)
			     .arg(startval+incval)
			     .arg(tl[3])
			     .arg(count>4?tl[4]:QString());
			 if(Debug>1){
			     ErrStream << "newfor="<<newfor<<"\n";
			     ErrStream << "newcmd="<<newcmd<<"\n";
			 }
			 CmdStack.prepend(newcmd);
			 CmdStack.prepend(newfor);
			 CmdStack.prepend(substituteVariables(newcmd));
		     }
		}else{  // for loop empty or finished
		        // No-op!
		}
	    }else{
		ErrStream <<tr("Invalid loop variable %1: %2\n")
		    .arg(var).arg(Command);
	    }

	}
    }else if(tl[0].lower()=="foreach"){ // !foreach <var> <v1> <v2> <v3> ...
	if(count>2){
	    QString var = tl[1];
	    if(validName.exactMatch(var)){
		 QString newcmd;
	         if(CmdStack.count()){
		     newcmd = CmdStack[0];
	             CmdStack.pop_front();
	         }else{
	             if(!CmdStream->atEnd()){
	                 newcmd = CmdStream->readLine();
	                 setLineNumber(CmdLine+1);
		     }
		 }
                 if( count > 2 ) {
		     if(!newcmd.isEmpty()){
			 Variables["{"+var+"}"] = tl[2];
			 if(count > 3){
		             QString newforeach = 
			         QString("!foreach %1").arg(var);
			     for(int i=3; i<count ; i++)
			         newforeach += " "+tl[i];
			     if(Debug>1){
			         ErrStream << "newforeach="<<newforeach<<"\n";
			         ErrStream << "newcmd="<<newcmd<<"\n";
			     }
			     CmdStack.prepend(newcmd);
			     CmdStack.prepend(newforeach);
                         }
			 CmdStack.prepend(substituteVariables(newcmd));
		     }
		}else{  // for loop empty or finished
		        // No-op!
		}
	    }else{
		ErrStream <<tr("Invalid loop variable %1: %2\n")
		    .arg(var).arg(Command);
	    }

	}
    }else if(tl[0].lower()=="prompt"){
        *OutStream <<cmd.mid(8);
	OutStream->device()->flush();
	QTextIStream Cin(stdin);
	QString answer("");
	char c;
	while(!Cin.atEnd()){
	    Cin.readRawBytes(&c,1);
	    if(c!='\r'){
		if(c=='\n')break;
		answer += c;
	    }
	}
	Variables["{ANSWER}"] = answer;
	if(Debug)ErrStream <<"ANSWER=("<<answer<<")\n";
    }else if(tl[0].lower()=="call"){   // !call <fct> <argval> ...
	evaluateFunction(tl);
    }else if(tl[0].lower()=="set"){
	if(count > 1)
	    setVariable(cmd.mid(5));
	else{
            QMap<QString,QString>::Iterator it;
            for ( it = Variables.begin(); it != Variables.end(); ++it ) 
		*OutStream << QString("%1=%2\n")
		    .arg(it.key().mid(1,it.key().length()-2)).arg(it.data());
	}

    }else if(tl[0].lower()=="unset"){
	unsetVariable(cmd.mid(7));
    }else if(tl[0].lower()=="print"){
        *OutStream <<cmd.mid(7)<<"\n";
	OutStream->device()->flush();
    }else{
	ErrStream <<tr("Unknown client command: %1\n").arg(Command);
    }

    return TRUE;
}

////////////////////////////////////////////////////////////////////////////

QString Client::getCommand()
{
    QString retval = Command;
    Command = QString();
    if(Debug>1)ErrStream <<tr("Cmd: %1\n").arg(retval);
    return retval;
}

////////////////////////////////////////////////////////////////////////////

void Client::resume()
{
    Waiting = FALSE;
    if(cmdAvailable())
        sendCommand();
    else
        if(QuitOnEof && Command.isEmpty() && !Waiting)
	    terminate();
}

////////////////////////////////////////////////////////////////////////////

QString Client::substituteVariables(const QString str)
{
    QString sub = str;
    QMap<QString,QString>::Iterator it;
    for ( it = Variables.begin(); it != Variables.end(); ++it ) 
        sub.replace(it.key(),it.data());
    return sub;
}

////////////////////////////////////////////////////////////////////////////

bool Client::setVariable(const QString& varspec)
{
    int sepeq = varspec.find('=');
    if(sepeq < 0 ) return FALSE;
    QString var = varspec.mid(0,sepeq).stripWhiteSpace();
    if(var.isEmpty())return FALSE;
    QString value = varspec.mid(sepeq+1).stripWhiteSpace();
    if(validName.exactMatch(var)){
        Variables[QString("{%1}").arg(var)] = value;
        return TRUE;
    }
    return FALSE;
}

////////////////////////////////////////////////////////////////////////////

bool Client::unsetVariable(const QString& varname)
{
    QString var = "{"+varname.stripWhiteSpace()+"}";
    if(Variables.contains(var)){
	Variables.remove(var);
	return TRUE;
    }
    return FALSE;
}

////////////////////////////////////////////////////////////////////////////

void Client::evaluateFunction(const QStringList& vl)
{
    int vcount = vl.count();
    if(vcount>1 && Functions.contains(vl[1])){
        if(Debug)ErrStream << tr("Calling function %1.\n").arg(vl[1]);
        QStringList function = Functions[vl[1]];
	int fcount = function.count();
	QStringList args = QStringList::split(' ',function[0],FALSE);
	int acount = args.count();
	QStringList values;
	for(int l = 1; l<fcount ; l++){
	    QString line = function[l];
	    if(line.find('{')>=0) for(int i = 1; i<acount ; i++){
	        line.replace("{"+args[i]+"}", i+1<vcount?vl[i+1]:QString(""));	
	    }
	    values << line;
	    if(Debug)ErrStream <<tr("line %1: %2\n").arg(l).arg(line);
	}
	CmdStack = values + CmdStack;
    }else{
        if(Verbose)ErrStream <<tr("Function %1 not defined!\n").arg(vl[1]);
    }
}



////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

Terminal::Terminal(QWidget *par, const QString& name)
    :QMainWindow(par,name)
{
    setMinimumWidth(fontMetrics().width("x")*72);
    setMinimumHeight(fontMetrics().height()*24);
    QMenuBar *menu = new QMenuBar(this);
    menu->insertItem("Close",this,SLOT(terminate()));
    QVBox *vb = new QVBox(this);
    setCentralWidget(vb);
    tv = new QTextView(vb,"tv");
    cb =  new QComboBox(TRUE,vb,"cb");
    cb->setFocus();
    connect(cb,SIGNAL(activated(const QString&)),this,SLOT(sendLine(const QString&)));
    show();
}

////////////////////////////////////////////////////////////////////////////

void Terminal::append(const QString& t)
{
    tv->append(QString(t).replace('\r',QString()));
    qApp->processEvents(0);
    tv->ensureVisible(0,tv->contentsHeight());
}

////////////////////////////////////////////////////////////////////////////

void Terminal::appendFormattedLine(const QString& t, const QString& font)
{
    QString text = t;
    text.replace('<',"&lt;");
    text.replace('>',"&gt;");
    if(!font.isEmpty())
	text = QString("<font %1> %2 </font>").arg(font).arg(text);
    tv->append(text);
}

////////////////////////////////////////////////////////////////////////////

void Terminal::sendLine(const QString& line)
{
    if(line.isEmpty())return;
    QString text = line+"\n";
    appendFormattedLine(text,"COLOR=#FF0000");
    emit textAdded(text);
    cb->setEditText("");
}

////////////////////////////////////////////////////////////////////////////

void Terminal::terminate()
{
    terminateAll();
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

int main( int argc, char **argv )
{
    QApplication a( argc, argv );
    argc = a.argc();
    argv = a.argv();

    // extract version from CVS keys
    Version = QString("$Name:  $").replace(QRegExp("_"),".");
    Version = cvsKey(Version);
    if(Version.isEmpty())Version = cvsKey("$Revision: 2.5 $");

    unsigned int delay = 10;

    OutStream = new QTextStream(stdout,IO_WriteOnly|IO_Translate);
    QFile OutputFile;

    QString EnifProgram;

    char *s;
    while ( --argc > 0 && (*++argv)[0] == '-')   /* check for options */
        for (s = argv[0] + 1; *s != '\0'; s++)
            switch (*s) {
            case 'h':           // help summary
                *OutStream << usage();
                return (0);
            case 'i':           // debug flag
                Interactive = TRUE;
                break;
            case 'v':           // debug flag
                Verbose = TRUE;
                break;
            case 'd':           // debug flag
                Debug++;
                break;
            case 'e':           // quit on error
                StopOnKO = FALSE;
                break;
            case 'L':           // specify program name to lauch
                if (!*++argv)
                    error(Client::tr("Missing argument: program name!"),ErrorMissingProgramName);
                EnifProgram = *argv;
                argc--;
		// no break here!!
            case 'l':           // launch an enif session
                LaunchEnif = TRUE;
		delay += 5000;
                break;
            case 'q':           // quit on error
                QuitOnEof = TRUE;
                break;
            case 'o':           // specify output file name
                if (!*++argv)
                    error(Client::tr("Missing argument: output file name!"),ErrorMissingOutputFile);
                OutputFile.setName(*argv);
		if(!OutputFile.open(IO_WriteOnly|IO_Translate))
                    error(Client::tr("Invalid output file name!"),
			    ErrorInvalidOutputFile);
                OutStream->setDevice(&OutputFile);
                argc--;
                break;
            case 's':           // specify host name of server 
                if (!*++argv)
                    error(Client::tr("Missing argument: host name!"),ErrorMissingHost);
                EnifHost = *argv;
                argc--;
                break;
            case 'p':           // contact Enif server at given port 
                if (!*++argv)
                    error(Client::tr("Missing argument: TCP port!"),ErrorMissingPort);
                bool ok;
	        EnifPort = QString(argv[0]).toInt(&ok);
		if(!ok)
                    error(Client::tr("Invalid TCP port %1!").arg(argv[0]),ErrorInvalidPort);
                argc--;
                break;
            default:
                error(Client::tr("Invalid option %1!\n\n%2").arg(*s).arg(usage()),ErrorInvalidOption);
            }
  
    if(LaunchEnif){
        if(EnifProgram.isNull())
	     EnifProgram = enifProgram();
	if(EnifPort==0){
	    EnifPort = searchAvailablePort(EnifHost,9090);
	    if(EnifPort==0) error(Client::tr("No available port found!")
		    .arg(argv[0]),ErrorNoAvailablePort);
	    else
		if(Verbose || Debug)
		    ErrStream <<Client::tr("Using port %1.\n").arg(EnifPort);
	}
	EnifProcess = new QProcess();
	EnifProcess->addArgument(EnifProgram);
	EnifProcess->addArgument("-q");
	EnifProcess->addArgument("-S");
	EnifProcess->addArgument(QString("%1").arg(EnifPort));
	if(!EnifProcess->start())
            error(Client::tr("Cannot launch Enif using %1!")
		    .arg(EnifProgram),ErrorLaunchEnif);
	else
	    if(Verbose)
		ErrStream <<Client::tr("Programm %1 launched.\n").arg(EnifProgram);
    }

    if(!EnifPort)EnifPort = 9090;

    MyClient = new Client(EnifHost,EnifPort,delay);

    if(EnifProcess){
        EnifProcess->connect(MyClient,SIGNAL(destroyed()),
	    EnifProcess,SLOT(kill()));
        EnifProcess->connect(EnifProcess,SIGNAL(processExited()),
	    qApp,SLOT(quit()));
    }
   
    if(argc>0){
        while(argv[0]){
	    MyClient->addArg(argv[0]);
	    argv++;argc--;
        }
	if(!Interactive)QuitOnEof = TRUE;
    }else{
	Interactive = TRUE;
    }

    if(Interactive){
        Terminal *term = new Terminal();
        MyClient->connect(MyClient,SIGNAL(receivedText(const QString&)),
            term,SLOT(append(const QString&)));
        MyClient->connect(
	    MyClient,SIGNAL(sentText(const QString&, const QString&)),
            term,SLOT(appendFormattedLine(const QString&, const QString&)));
        MyClient->connect(term,SIGNAL(textAdded(const QString&)),
            MyClient,SLOT(insertCommand(const QString&)));
        term->show();
    }

    return a.exec();

}

////////////////////////////////////////////////////////////////////////////

void error(const QString& msg, int errorcode)
{
    ErrStream <<msg<<"\n";
    terminateAll(errorcode);
}

////////////////////////////////////////////////////////////////////////////

int searchAvailablePort(const QString& host, int startport)
{
    (void) host;
    QFile useport("useport");
    int portno = startport;
    if(useport.exists() && useport.open(IO_ReadWrite|IO_Translate)){
	 useport.at(0);
	 QTextStream ps(&useport);
	 ps >> portno;
	 if(portno > 0){
	     useport.flush();
	     useport.at(0);
	     ps <<portno+1;
	     useport.close();
	 }
    }
    if(!portno) portno = startport;
    if(!portno) portno = 9090;
    return portno;
}

////////////////////////////////////////////////////////////////////////////

void terminateAll(int ec)
{
    static bool delayed = FALSE;
    if(delayed)return;
    if(EnifProcess){
	 delayed = TRUE;
	 if(Verbose)ErrStream <<Client::tr("Terminate Enif session.\n");
         EnifProcess->tryTerminate();
         QTimer::singleShot( 5000, EnifProcess, SLOT( kill() ) );
         QTimer::singleShot( 7000, qApp, SLOT( quit() ) );
    }else
        qApp->exit(ec);
}

//////////////////////////////////////////////////////////////////////////////

QString cvsKey(const QString& key)
{
    if(key.left(1)!="$" || key.right(1)!="$")return key;
    int icol = key.find(':');
    if(icol<0)return key.mid(1,key.length()-2);
    return key.mid(icol+1,key.length()-icol-2).stripWhiteSpace();
}

QString enifProgram()
{
    QString enifprog;
#ifdef Q_OS_WIN32
    enifprog = "enif.exe";
#else
    enifprog = "enif";
#endif
    QFileInfo fi;
    QString fn;
    if(getenv("ENIFPATH")){
	fn = QString(getenv("ENIFPATH"))+"/bin/"+enifprog;
	if(Debug>1)ErrStream <<Client::tr("Trying %1...\n").arg(fn);
	fi = QFileInfo(fn);
	if(fi.exists() && fi.isExecutable())return fn;
	fn = QString(getenv("ENIFPATH"))+"/"+enifprog;
	if(Debug>1)ErrStream <<Client::tr("Trying %1...\n").arg(fn);
	fi = QFileInfo(fn);
	if(fi.exists() && fi.isExecutable())return fn;
    }
    if(getenv("EMME2")){
	fn = QString(getenv("EMME2"))+"/enif/"+enifprog;
	if(Debug>1)ErrStream <<Client::tr("Trying %1...\n").arg(fn);
	fi = QFileInfo(fn);
	if(fi.exists() && fi.isExecutable())return fn;
    }
    return enifprog;
}

//////////////////////////////////////////////////////////////////////////////
//
// usage() - display usage information including calling sequence and options
//

QString usage()
{
    return 
Client::tr("SEnC - Simple Enif Client (%1)  -  Heinz Spiess, 2003-05-26\n\n")
   .arg(Version) +

Client::tr("Calling sequence:  senc [options] [file|varspec ...]\n\n") +

Client::tr("Available options:\n") +
Client::tr("        -d                  Generate debugging output\n") +  
Client::tr("        -e                  Continue with input processing after errors\n") +  
Client::tr("        -h                  Display this help summary\n") +  
Client::tr("        -i                  Interactive mode\n") +  
Client::tr("        -l                  Launch enif\n") +  
Client::tr("        -L <progname>       Launch enif using specified program name\n") +
Client::tr("        -o <outputfile>     Send the standard output to this file.\n") +
Client::tr("        -p <port>           Port used by enif server\n") +
Client::tr("        -q                  Quit after all input files are processed\n") +  
Client::tr("        -s <server>         Name or IP of host running enif server\n") +
Client::tr("        -v                  Generate verbose output\n") +
QString("\n") +
Client::tr("Client commands:\n") +
Client::tr(" !call <fct> [args...]       Call a function\n") +
Client::tr(" !for <var> <v1> <v2> [inc]  Run for-loop on next command\n") +
Client::tr(" !foreach <var> <v1> <v2> .. Run foreach-loop on next command\n") +
Client::tr(" !function <fct> [arg ...]   Begin of function definition\n") +
Client::tr(" !!                          End of function definition\n") +
Client::tr(" !if <v1> [<comp> <v2>]      Conditional execution of next command\n") +
Client::tr(" !iferror                    Execute next command only if an error occurred\n") +
Client::tr(" !print <text>               Print text to standard output\n") +
Client::tr(" !prompt <text>              Prompt and wait for user input\n") +
Client::tr(" !set <variable>=<value>     Set a variable to specified value\n") +
Client::tr(" !unset <variable>           Unset a variable, i.e. delete it\n") +
Client::tr(" !wait <seconds>             Pause for the specified number of seconds\n") +
QString("\n");
}
