// senc.h

#ifndef SENC_H
#define SENC_H

// $Source: /u/cvs/senc/senc.h,v $
// $Revision: 2.3 $
// $Author: heinz $
// $Date: 2003/05/29 09:17:32 $
// $Name:  $
// $State: Exp $

/////////////////////////////////////////////////////////////////////////////
//
// This program has been written by
//
//        Heinz Spiess, EMME/2 Support Center, CH-2558 Aegerten, Switzerland
//        heinz@spiess.ch
//
// 
// Compiled binaries of this program can be distributed under the following
// conditions:
//
//  1) This program, as well as programs derived from it, cannot be sold
//     for a fee (but it can be used and distributed without charge within
//     consulting projects).
//
//  2) The user is aware that this program is not a part of the EMME/2 .
//     Enid software licence and there is no explicit or implied warranty
//     or support provided with this program.
//
// In addition to the above, the author grants the right to access and
// using the source code of this program as an example application for
// writing Enif client programs.  Integrating parts of the source code into
// other programs or making modifications to the programs is permitted
// under the following additional conditions:
//
//  3) Anyone who changes, enhances or otherwise modifies the source code
//     of SEnC is obliged to rename the changed source files and to use
//     a different program name for the modified program.
//
//  4) The source code of such derived programs must contain a reference
//     to the SEnC program and the particular version which it based on.
//
//                                              Heinz Spiess,
//                                              Aegerten, 24. Mai 2003
//
/////////////////////////////////////////////////////////////////////////////

// $Log: senc.h,v $
// Revision 2.3  2003/05/29 09:17:32  heinz
// improved search for location of Enif binary with option -l
//
// Revision 2.2  2003/05/28 18:35:31  heinz
// the version identifier is now extracted from the CVS keys
//
// Revision 2.1  2003/05/28 18:27:42  heinz
// version corresponding the revised SEnC paper dated May 2003
//

#include <qobject.h>
#include <qfile.h>
#include <qmap.h>
#include <qtimer.h>
#include <qregexp.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qtextview.h>
#include <qtextstream.h>
#include <qmainwindow.h>
#include <qcombobox.h>

class QSocket;
class QTextStream;
class QTextIStream;

class Client : public QObject
{
    Q_OBJECT
    QMap<QString,QStringList> Functions;
    QMap<QString,QString> Variables;
    QStringList Args;
    QStringList CmdStack;
    QString Command;
    QTextStream * CmdStream;
    QFile CmdFile;
    int CmdLine;
    QTimer Timer;
    static QRegExp validName;
    int MaxStartupDelay;
    int DelayIncrement;

    unsigned int Port;
    QString Host;
    QSocket *Socket;
    QTextStream * TextStream;

    bool cmdAvailable();
    void checkCommand();
    QString getCommand();
    bool clientCommand(const QString&);
    void evaluateFunction(const QStringList&);
    bool setVariable(const QString& varspec);
    bool unsetVariable(const QString& varname);
    QString substituteVariables(const QString str);
    void wait(float delay);
    void setLineNumber(int lineno);

    bool Waiting;
    bool Stopped;
    bool KOReceived;
    
public:
    Client( const QString &host, unsigned int port , unsigned int delay = 0);
    ~Client();
    QSocket *socket() const {return Socket;}
    QTextStream *textStream() const {return TextStream;}
    void addArg(const QString&);
    void setArgs(const QStringList&);
    QStringList args() const {return Args;}

public slots:
    void insertCommand( const QString & );
    void sendText( const QString & );
    void resume();

signals:
    void finished();
    void status( const QString & );
    void receivedText( const QString &);
    void sentText( const QString &t, const QString &f);

private slots:
    void readyRead();
    void connected();
    void deleteMe();
    void socket_error(int);
    void hostFound();
    bool sendCommand();
    void connectToHost();
    void terminate();
};

class Terminal : public QMainWindow
{
    Q_OBJECT
    QTextView *tv;
    QComboBox *cb;

public:
    Terminal(QWidget *parent=0, const QString& name=0);
public slots:
    void append(const QString& t);
    void appendFormattedLine(const QString& t, const QString& font = 0);
    void sendLine(const QString&);
    void terminate();
signals:
    void textAdded(const QString&);
};

enum ErrorCodes {
    NoError = 0,
    ErrorMissingHost,
    ErrorMissingPort,
    ErrorMissingProgramName,
    ErrorMissingOutputFile,
    ErrorInvalidOutputFile,
    ErrorInvalidPort,
    ErrorLaunchEnif,
    ErrorNoAvailablePort,
    ErrorInvalidOption
};

QString usage();
void error(const QString&, int);
int searchAvailablePort(const QString& host, int startport);
void terminateAll(int = 0);
QString cvsKey(const QString&);
QString enifProgram();

#endif //SENC_H
